/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "ConstraintDB.hpp"
#include "Settings.hpp"


void ConstraintDB::clsCreateBinary(const Literal lit0, const Literal lit1){

    assert(lit0.idx() < watches.size() && lit1.idx() < watches.size());

    vector<WatchedElement>& l0_watches = watches[lit0.idx()];
    vector<WatchedElement>& l1_watches = watches[lit1.idx()];

    //find subsumed ternary clauses
    vector<Literal> subsumed_ternaries;
    vector<WatchedElement>::iterator iter, to;
    for(iter = to = l0_watches.begin();iter != l0_watches.end();++iter){
    	if(iter->isTernary()){
    		if(iter->getFirstLiteral() == lit1){
    			subsumed_ternaries.push_back(iter->getSecondLiteral()); continue;}
    		if(iter->getSecondLiteral() == lit1){
    			subsumed_ternaries.push_back(iter->getFirstLiteral());	continue;}
    	}
    	*to++ = *iter;
    }

    //some ternary clauses are subsumed
    if(to != iter){
    	l0_watches.erase(to,l0_watches.end());
    	uint subsumed = subsumed_ternaries.size();
    	statistics.nbr_bin_subsumes_tern += subsumed;

    	nbr_tern_clses 		-= subsumed;
    	nbr_short_literals	-= 3*subsumed;

    	for(iter = to = l1_watches.begin();iter != l1_watches.end();++iter){
    		if(!iter->isTernary() || (iter->getFirstLiteral() != lit0 && iter->getSecondLiteral() != lit0)){
    			*to++ = *iter;
    		}
    	}
    	assert(l1_watches.end() - to == (int)subsumed);
    	l1_watches.erase(to,l1_watches.end());

    	//clear ternaries
    	for(uint i = subsumed; i--;){
    		Literal lit2 = subsumed_ternaries[i];
    		vector<WatchedElement>& l2_watches = watches[lit2.idx()];

    		for(iter = to = l2_watches.begin();iter != l2_watches.end();++iter){
    		    if(iter->isTernary()){
    		    	if(	(iter->getFirstLiteral() == lit0 && iter->getSecondLiteral() == lit1) ||
    		    		(iter->getFirstLiteral() == lit1 && iter->getSecondLiteral() == lit0)) continue;
    		    }
    		    *to++ = *iter;

    		}
    		l2_watches.erase(to,l2_watches.end());
    	}
    }


    l0_watches	.push_back(WatchedElement(lit1));
   	l1_watches	.push_back(WatchedElement(lit0));
   	nbr_short_literals	+= 2;
   	++nbr_bin_clses;
}

bool ConstraintDB::clsFindBinary(const Literal lit0, const Literal lit1){

	assert(lit0.idx() < watches.size() && lit1.idx() < watches.size());

    vector<WatchedElement>& l0_watches = watches[lit0.idx()];
	for(vector<WatchedElement>::iterator iter = l0_watches.begin();iter != l0_watches.end();++iter){
		if(iter->isBinary() && iter->getFirstLiteral() == lit1) return true;
	}

	return false;
}

bool ConstraintDB::clsFindTernary(const Literal lit0, const Literal lit1, const Literal lit2 ){

	assert(lit0.idx() < watches.size() && lit1.idx() < watches.size() && lit2.idx() <watches.size());

	vector<WatchedElement>& l0_watches = watches[lit0.idx()];
	for(vector<WatchedElement>::iterator iter = l0_watches.begin();iter != l0_watches.end();++iter){
		if(iter->isBinary() && (iter->getFirstLiteral() == lit1 || iter->getFirstLiteral()==lit2)) return true;
		if(iter->isTernary()){
			if((iter->getFirstLiteral() == lit1 && iter->getSecondLiteral() == lit2)
			|| (iter->getFirstLiteral() == lit2 && iter->getSecondLiteral() == lit1)) return true;
		}
	}

	return clsFindBinary(lit1,lit2);
}



void ConstraintDB::clsCreateTernary(const Literal lit0, const Literal lit1,const Literal lit2){

    assert(lit0.idx() < watches.size() && lit1.idx() < watches.size() && lit2.idx() < watches.size());

	watches[lit0.idx()]	.push_back(WatchedElement(lit1,lit2));
	watches[lit1.idx()]	.push_back(WatchedElement(lit0,lit2));
	watches[lit2.idx()]	.push_back(WatchedElement(lit0,lit1));
	nbr_short_literals	+= 3;
	++nbr_tern_clses;
}


Clause* ConstraintDB::clsCreate(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool learnt,bool cardinality){

    uint32_t cl_sz							= lend-lbeg;
    assert(cl_sz > 3);

    void* mem								= malloc(sizeof(Clause) + sizeof(Literal)*cl_sz);
    if(mem == NULL){
    	print_out("\nc !!!OUT OF MEMORY !!!\n");
    	exit(EXIT_FAILURE);
    	return NULL;
    }

    Clause* cls 							= new (mem) Clause(cl_sz, learnt,cardinality);

    Literal lit0 							= *lbeg;
    Literal lit1 							= *(lbeg+1);

    for(uint32_t i = 0;lbeg != lend;++lbeg) {(*cls)[i++] = *lbeg;}

    if(learnt)  {
        learnt_cls                          .push_back(cls);
        nbr_long_learnt_literals			+= cl_sz;
    }else{
        fix_cls                             .push_back(cls);
        nbr_long_fix_literals				+= cl_sz;
    }

    clsLinkWatched(cls);

    return cls;
}


void    ConstraintDB::clsUncreate(Clause* cl,bool unlink){
    assert(cl->size() > 1);

    if(cl->isLearnt())	nbr_long_learnt_literals 	-= cl->size();
    else				nbr_long_fix_literals 		-= cl->size();

    if(unlink) {clsUnlinkWatched(cl);}

    free(cl);
}




void ConstraintDB::reserveForVariables(uint32_t need_sz){

    if(need_sz > assignments.size()){
        //Variable infos
        assignments     .resize(need_sz,UNKNOWN_TrBool.toInt());
        assign_info     .resize(need_sz,AssignInfo());

        //Literal infos
        need_sz = need_sz << 1;
        watches         .resize(need_sz);
    }
}




void ConstraintDB::learntsGC(uint32_t new_sz){

	if(new_sz >= learnt_cls.size()){return;}

    statistics.nbr_learnt_reductions += 1;

    std::nth_element(learnt_cls.begin(), learnt_cls.begin() + new_sz - 1,
                       learnt_cls.end(), Clause::RelevanceDescendingFunctor());

    for(uint32_t del_idx = new_sz; del_idx < learnt_cls.size();++del_idx){
    	assert(!learnt_cls[del_idx]->isCardConstraint());
    	if(clsLocked(*(learnt_cls[del_idx])))   learnt_cls[new_sz++] = learnt_cls[del_idx];
    	else                                    clsUncreate(learnt_cls[del_idx]);
    }

    learnt_cls.resize(new_sz);
}



bool ConstraintDB::clsCanBeOmitted(Clause& cl){


	for(uint32_t i = 0; i != cl.size(); ) {
		if(litIsTrue(cl[i])){ 							return true;}
		if(litIsFalse(cl[i]) && i > 1){ // if i <= 1 -> clause is unit and other watcher should be true
			clsRemoveLiteral(cl,i);
			continue;
		}
		++i;
	}

	const uint32_t sz = cl.size(); assert(sz > 1);

	if(sz > 3) 											return false;

	if(sz == 3){clsCreateTernary(cl[0],cl[1],cl[2]); 	return true;}

	clsCreateBinary(cl[0],cl[1]);						return true;
}


void ConstraintDB::removeSat(vector<Clause*>& cls){
    vector<Clause*>::iterator to, iter;
    for (iter = to = cls.begin(); iter != cls.end(); ++iter){
        if (clsCanBeOmitted(**iter)){
        	statistics.nbr_simplify_worked += 1;
            clsUncreate(*iter);
        }else{
        	assert(litIsUnknown((**iter)[0])&& litIsUnknown((**iter)[1]));
            *to++ = *iter;
        }
    }
    cls.erase(to,cls.end());
}




bool ConstraintDB::debugVerifyModel(){


	for(int i(fix_cls.size());i--;){
        if(checkSatisfied(*(fix_cls[i]))) continue;

        if(console_out != NULL){
        	console_out->writeString("|Found unsatisfied clause: ");
            Clause& cl = *(fix_cls[i]);
            for(uint32_t j = 0; j != cl.size(); ++j) {
                Literal lit = cl[j];
                console_out->writeString(" %i: %s\n",lit.toDIMACS(),(litIsFalse(lit)?"false":"unknown"));
            }
        }

        return false;
    }

    for(uint i = 0; i != numberVariables();++i){
    		Variable var(i);
    		assert(!litIsUnknown(var.toPosLit()) || litWatchedClauses(var.toPosLit()).empty() && litWatchedClauses(var.toNegLit()).empty() );

    		Literal true_lit = litIsTrue(var.toPosLit()) ? var.toPosLit() : var.toNegLit();
    		we_vec& watches = litWatchedClauses(true_lit.flipped());

    		for(uint j = 0; j != watches.size(); ++j){
    			WatchedElement& we = watches[j];
    			if(!we.isShort() || litIsTrue(we.getFirstLiteral()) || (we.isTernary() && litIsTrue(we.getSecondLiteral()))) continue;
    			if(console_out != NULL){
    				console_out->writeString("|Found unsatisfied short clause: ");
    				console_out->writeString(" %i: %s\n",true_lit.flipped().toDIMACS(),(litIsFalse(true_lit.flipped())?"false":"unknown"));
    				console_out->writeString(" %i: %s\n",we.getFirstLiteral().toDIMACS(),(litIsFalse(we.getFirstLiteral())?"false":"unknown"));
    			    if(we.isTernary()){
    			    	console_out->writeString(" %i: %s\n",we.getSecondLiteral().toDIMACS(),(litIsFalse(we.getSecondLiteral())?"false":"unknown"));}
    			}
    			return false;
    		}
   }

    if(console_out != NULL)
    	console_out->writeString("|Verified %d original clauses and %d short clauses ",fix_cls.size(), numberShortClses());
    return true;
}



bool ConstraintDB::checkAssignReason(){

    for(uint32_t i(assign_info.size());i--;){
        if(assign_info[i].noReason() || assign_info[i].isBinReason()) continue;

        assert(assign_info[i].getClsReason() != NULL);
        Clause& c = *(assign_info[i].getClsReason());
        Variable v(i);
        if(c[0].toVar() != v && c[1].toVar() != v){
            print_out("Variable %u has invalid assign reason!",v.toDIMACS());
            return false;
        }
    }

    return true;
}

void ConstraintDB::printModel(){
	print_out("variables assigned true ");
	const uint32_t sz = numberVariables();
	for(uint i = 0; i != sz;++i){
		Variable v(i);
		TrBool val = varValue(v);
		//assert(val.isKnown());
		if(val.isTrue()){print_out("%d ",v.toDIMACS());}
	}
	print_out(" 0\n");
}



bool ConstraintDB::checkSatisfied(Clause& cl) const{
    for(uint32_t i = 0; i != cl.size(); ++i) {
        if(litIsTrue(cl.literals[i])) return true;
    }
    return false;
}


bool ConstraintDB::debugCheckCardWatchers(we_vec& wes){
	for(we_iter iter = wes.begin();iter != wes.end();++iter){
		if(iter->isCardinality()){
			Clause* cl = iter->getClause();
			assert(cl != NULL);
			if(!cl->isCardConstraint()){
				print_out("Bad card at clause \n");
				return false;
			}
		}
	}
	//print_out("Check for one vec with %d elements ok\n",wes.size());
	return true;
}

bool ConstraintDB::debugCheckCardWatchers(){
	for(uint i = watches.size();i--;){
		if(!debugCheckCardWatchers(watches[i])) return false;
	}
	//print_out("Card watchers ok \n");
	return true;
}


const char*  constraintdb_cpp_version(){return "ConstraintDB.cpp:   $Revision: 1.7 $";}


