/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 based on the heap implementation of Minisat 20
 */

#ifndef DECAYHEAP_HPP_
#define DECAYHEAP_HPP_

#include<vector>
#include<cassert>
#include<stdint.h>


template<class T>
class INTMap{
	vector<uint32_t>		  values;

public:

	void reserve(uint sz,int def){values.resize(sz,def);}

	uint32_t get(const T& a){
		assert(a.idx() < values.size());
		return values[a.idx()];
	}

	void set(const T& a, uint32_t idx){
		assert(a.idx() < values.size());
		values[a.idx()] = idx;
	}
};


template<class T, class A>
class ActivityMap{

	A max;
	A min;
	A increment;
	A decay;

	vector<A> activity;

public:
	ActivityMap(A MAX,A MIN, A dec):
		max(MAX),min(MIN),increment(MIN),decay(dec){};

	void reserve(uint sz){ activity.resize(sz,A(0));}

	bool less(const T& a, const T& b){
		assert(a.idx() < activity.size() && b.idx() < activity.size());
		return activity[a.idx()] < activity[b.idx()];
	}

	void decayAll(){increment *= decay; };

	bool bump(const T& v){

		activity[v.idx()] += increment;

	    if(activity[v.idx()] > max){
	    	const var_act_t::denominator_t w = min.getLogOfDenominator();
	        increment.weaken(w);
	        int i = activity.size();
	        while(i--) activity[i].weaken(w);
	        return false;
	    }
	    return true;
	}
};


/**
 * Heap structure that ranks all elements according to some activity number.
 * It allows for incrementation of the activity number of single elements and moreover, all activities can be decremented at once.
 * KeyType  - has to implement method 'int idx()' and has to be constructable by 'KeyType(int index)'
 * ValueType - type to rank elements has to be comparable by operators > and >= value has to be increasable by operator += */
template<class KeyType, class Comp , class IntMap = INTMap<KeyType> >
class MaxHeap{



	IntMap* intMap;
	Comp*	comp;

	static inline uint32_t   parent		(uint32_t index)	 {return (index-1) >> 1;};
	static inline uint32_t   leftChild	(uint32_t index)	 {return (index << 1) + 1;};
	static inline uint32_t   rightChild	(uint32_t index)	 {return (index << 1) + 2;};

	/**The actual heap. Each element of the heap vector is an index of the original element.*/
	std::vector<KeyType> heap;

protected:
	inline uint32_t	moveUp(const KeyType a, uint32_t heap_index);
	inline void		moveDown(const KeyType a, uint32_t heap_index);

	/**Returns element id of the element that was at the specified position pos.
	 * Does not ensure the heap property*/
	inline KeyType	removeAtPosNoUpdate(uint32_t pos);


	inline bool		checkHeapProperty(uint32_t heap_index);

public:
	static const uint32_t HEAP_INVALID_INDEX = ~0;

	MaxHeap(Comp* c,IntMap* a):intMap(a),comp(c){ };

	MaxHeap(const MaxHeap<KeyType,IntMap,Comp >& h)
		:intMap(h.intMap),comp(h.comp){assert(h.heap.empty());};

	~MaxHeap() { };

	inline KeyType	removeAtPos(uint32_t i);

	inline void		remove(const KeyType& element);

	inline KeyType	removeRandom(Random& randy);

	inline KeyType	removeMax();

	inline void		removeAll();

	bool			contains(const KeyType& element){
					return intMap->get(element) != HEAP_INVALID_INDEX;}

	/**Insert a new element in the heap.
	 * @param element: a reference to the element to be inserted
	 * @param initial_value: the initial rank value that is 0 per default.*/
	inline void		addNewElementAndUpdateValue(const KeyType& element);

	/**Insert a new element in the heap. This function won't arrange the heap property itself.
	* It has to be ensured by the user that the heap property holds, for instance if the initial_value is not greater than the one of all
	* previously added elements.
	* @param element: a reference to the element to be inserted
	* @param initial_value: the initial rank value that is 0 per default.*/
	inline void		addElementIfNotExists(const KeyType& element);

	/**Reactivates an element that was previously removed by the function removeMax or removeAtPos.*/
	inline void		restoreElement(const KeyType& element);

	inline void 	valueChanged(const KeyType& element);

	inline void 	valueIncreased(const KeyType& element);

	/**Return the size of the heap*/
	uint32_t		size()				  	{return heap.size();};

	/**Checks if the heap is empty*/
	bool		 	empty()				 	{return heap.empty();}

	KeyType	  		operator[](uint32_t i)	{assert(i < heap.size()); return KeyType(heap[i]);}

};





//Inline Implementation


template<class KeyType, class IntMap, class Comp>
uint32_t  MaxHeap<KeyType,IntMap,Comp>::moveUp(const KeyType element, uint32_t heap_index){
	assert(heap_index < heap.size());
	assert(heap[heap_index] == element);
	uint32_t parent_index;
	while(  heap_index != 0
		&& (parent_index = parent(heap_index), comp->less(heap[parent_index],element))){

		heap[heap_index]								=   heap[parent_index];
		intMap->set(heap[heap_index],heap_index);
		heap_index										=   parent_index;
	}
	heap[heap_index]									=   element;
	intMap->set(element,heap_index);
	return heap_index;
}

template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::moveDown(const KeyType element, uint32_t heap_index){
	assert( heap_index < heap.size() && heap[heap_index] == element);
	uint32_t child_index;
	while((child_index = leftChild(heap_index), child_index < heap.size()))
	{
		//get the child with the higher rank
		if(child_index+1 < heap.size() && comp->less(heap[child_index],heap[child_index+1]))
			++child_index; //the right one is bigger


		if(comp->less(heap[child_index],element))
			break;

		heap[heap_index]								=   heap[child_index];
		intMap->set(heap[heap_index],heap_index);
		heap_index										=   child_index;
	}
	heap[heap_index]									=   element;
	intMap->set(element,heap_index);
}




//has to be ensured that not empty
template<class KeyType, class IntMap, class Comp>
KeyType MaxHeap<KeyType,IntMap,Comp >::removeAtPosNoUpdate(uint32_t pos){

	assert(pos != HEAP_INVALID_INDEX);
	assert(pos < heap.size());

	KeyType e							= heap[pos];
	KeyType last						= heap.back();
	heap[pos]							= last;
	intMap->set(last,pos);
	intMap->set(e,HEAP_INVALID_INDEX);//we leave value unchanged for re-adding
	heap.pop_back();

	return e;
}

template<class KeyType, class IntMap, class Comp>
KeyType MaxHeap<KeyType,IntMap,Comp >::removeAtPos(uint32_t pos){
	assert(pos != HEAP_INVALID_INDEX);
	KeyType ret = removeAtPosNoUpdate(pos);
	if(heap.size() > 1 && pos < heap.size()){
		pos = moveUp(heap[pos],pos);
		moveDown(heap[pos],pos);
	}
	costly_assert(checkHeapProperty(0));
	return ret;
}


template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::remove(const KeyType& element){
	uint32_t pos = intMap->get(element);
	if(pos == HEAP_INVALID_INDEX) return;
	removeAtPos(pos);
}


template<class KeyType, class IntMap, class Comp>
KeyType MaxHeap<KeyType,IntMap,Comp >::removeMax(){
	KeyType ret = removeAtPosNoUpdate(0);
	if(heap.size() > 1)  moveDown(heap[0],0);

	costly_assert(checkHeapProperty(0));
	return ret;
}

template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::removeAll(){

	for(uint i = heap.size();i--;){
		intMap->set(heap[i],HEAP_INVALID_INDEX);
	}
	heap.clear();
}



template<class KeyType, class IntMap, class Comp>
KeyType MaxHeap<KeyType,IntMap,Comp >::removeRandom(Random& randy){
	return removeAtPos(randy.nextInt(heap.size()));
}


template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::valueChanged(const KeyType& element){

	uint32_t pos = IntMap(element);
	if(pos != HEAP_INVALID_INDEX){
		assert(heap[pos]==element);
		pos = moveUp(element,pos);
		moveDown(heap[pos],pos);
	}
}


template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::valueIncreased(const KeyType& element){

	uint32_t pos = intMap->get(element);
	if(pos != HEAP_INVALID_INDEX){
		assert(heap[pos] == element);
		moveUp(element,pos);
	}
}


template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::addElementIfNotExists(const KeyType& element){

	if(intMap->get(element) != HEAP_INVALID_INDEX) return;

	int sz = heap.size();
	heap.push_back(element);
	intMap->set(element,sz);

}


template<class KeyType, class IntMap, class Comp>
void MaxHeap<KeyType,IntMap,Comp >::restoreElement(const KeyType& element){

	if(intMap->get(element) != HEAP_INVALID_INDEX) return;

	int sz = heap.size();
	intMap->set(element,sz);
	heap.push_back(element);
	assert(heap[sz]==element);

	moveUp(element,sz);
}


template<class KeyType, class IntMap, class Comp>
bool MaxHeap<KeyType,IntMap,Comp >::checkHeapProperty(uint i){
	return i >= heap.size()
		  || ((i == 0 || !comp(heap[parent(i)],heap[i])) && checkHeapProperty(leftChild(i)) && checkHeapProperty(rightChild(i)));
}

#define decay_heap_hpp_version "DecayHeap.hpp:	  $Revision: 1.2 $"


#endif /*DECAYHEAP_HPP_*/


